#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <str/str.h>
#include <msg/msg.h>
#include <net/socket.h>
#include "mailfront.h"

#define POLICYD_MODE_LOCAL 1
#define POLICYD_MODE_TCP 2

#define POLICYD_DEFER_MSG  "Message delivery deferred by server policy. Please try again later."
#define POLICYD_REJECT_MSG "Message delivery rejected by server policy."

static int           policyd_lookup_mode;
static str           policyd_path;
static ipv4addr      policyd_addr;
static ipv4port      policyd_port;
static unsigned long policyd_timeout;
static unsigned long policyd_connect_timeout;
static unsigned long policyd_send_timeout;
static unsigned long policyd_failsafe;
static str           policyd_helo;
static str           policyd_sender;
static str           line;
static int           have_sender;

static RESPONSE(nohelo, 503, "5.5.1 Please say EHLO or HELO first.");
static RESPONSE(nosender, 503, "5.5.1 You must send MAIL FROM: first.");
static response resp_policyd = { 0, 0 };


static const response *warn_conf(void)
{
  warn1("policyd skipped: broken setup");
  return policyd_failsafe ? 0 : &resp_internal;
}

static const response *warn_sys(const char *s1)
{
  warn2sys("policyd skipped: ", s1);
  return policyd_failsafe ? 0 : &resp_internal;
}

static const response *init(void)
{
  const char *policyd_lookup;
  const char *tmp;
  size_t len;

  /* environment variable POLICYD_LOOKUP must be set */
  if ((policyd_lookup = getenv("POLICYD_LOOKUP")) == 0) return 0;
  /* environment variable POLICYD_LOOKUP must not be empty */
  if (*policyd_lookup == 0) return 0;
  /* environment variable POLICYD_LOOKUP must contain at least on ':' */
  if ((tmp = strchr(policyd_lookup, ':')) == 0)
    return warn_conf();
  /* environment variable POLICYD_LOOKUP must not start with ':' */
  if ((len = tmp - policyd_lookup) == 0)
    return warn_conf();
  ++tmp;
  /* environment variable POLICYD_LOOKUP must not end with ':' */
  if (*tmp == 0)
    return warn_conf();
  /* check for contact method, copy address */
  if (strncasecmp(policyd_lookup, "local", len) == 0) {
    policyd_lookup_mode = POLICYD_MODE_LOCAL;
    str_copys(&policyd_path, tmp);
  }
  else if (strncasecmp(policyd_lookup, "tcp", len) == 0) {
    /* for tcp mode address must have format address:port */
    if ((tmp = ipv4_scan(tmp, &policyd_addr)) == 0)
      return warn_conf();
    if (*tmp != ':') return warn_conf();
    policyd_port = strtoul(tmp + 1, (char **)&tmp, 10);
    if (*tmp != 0) return warn_conf();
    policyd_lookup_mode = POLICYD_MODE_TCP;
  }
  if ((tmp = getenv("POLICYD_FAILSAFE")) && *tmp)
    policyd_failsafe = strtoul(tmp, 0, 10);

  /* timeouts for policyd - defaults to 5000 ms */
  if (
    (tmp = getenv("POLICYD_TIMEOUT")) == 0 ||
    (policyd_timeout = strtoul(tmp, (char**)&tmp, 10)) == 0 ||
    *tmp != 0
  )
    policyd_timeout = 5000;
  if (
    (tmp = getenv("POLICYD_CONNECT_TIMEOUT")) == 0 ||
    (policyd_connect_timeout = strtoul(tmp, (char**)&tmp, 10)) == 0 ||
    *tmp != 0
  )
    policyd_connect_timeout = policyd_timeout;
  if (
    (tmp = getenv("POLICYD_SEND_TIMEOUT")) == 0 ||
    (policyd_send_timeout = strtoul(tmp, (char**)&tmp, 10)) == 0 ||
    *tmp != 0
  )
    policyd_send_timeout = policyd_timeout;

  return 0;
}

static const response *helo(str* hostname)
{
  if (!policyd_lookup_mode) return 0;
  str_copy(&policyd_helo, hostname);
  str_copys(&policyd_sender, "");	/* forget any previous sender */
  have_sender = 0;
  return 0;
}

static const response *reset(void)
{
  if (!policyd_lookup_mode) return 0;
//  str_copys(&policyd_helo, "");		/* forget any previou helo host name */
  str_copys(&policyd_sender, "");	/* forget any previous sender */
  have_sender = 0;
  return 0;
}

static const response *sender(str* email_addr)
{
  if (!policyd_lookup_mode) return 0;
  str_copy(&policyd_sender, email_addr);
  have_sender = 1;
  return 0;
}

static const response *recipient(str* email_addr)
{
  const response *r = 0;
  const char *client_address;
  const char *action;
  char *tmp;
  int sock;
  int x;
  size_t size;
  unsigned long int ul;
  ibuf netin;
  obuf netout;

  if (!policyd_lookup_mode) return 0;
  client_address = getprotoenv("REMOTEIP");
  /* must never happen under tcpserver, works with sslserver (PROTO=="SSL") */
  if (client_address == 0 || *client_address == 0) return 0;

  if (policyd_helo.len == 0) return &resp_nohelo;
  if (!have_sender) return &resp_nosender;

  /* create a socket and connect it to policy daemon */
  switch (policyd_lookup_mode) {
  case POLICYD_MODE_LOCAL:
    if ((sock = socket_unixstr()) < 0)
      return warn_sys("unable to create socket for policyd");
    x = socket_connectu_timeout(sock, policyd_path.s, policyd_connect_timeout);
    break;
  case POLICYD_MODE_TCP:
    if ((sock = socket_tcp()) < 0)
      return warn_sys("unable to create socket for policyd");
    x = socket_connect4_timeout(sock, &policyd_addr, policyd_port, policyd_connect_timeout);
    break;
  default:
    return 0;
  }
  if (!x) {
    r = warn_sys("unable to connect to policyd");
    close(sock);
    return r;
  }
  if (ibuf_init(&netin, sock, 0, 0, 4096)) {
    netin.io.timeout = policyd_timeout;
    if (obuf_init(&netout, sock, 0, 0, 4096)) {
      netout.io.timeout = policyd_send_timeout;
      if (
	obuf_write(&netout,
	  "request=smtpd_access_policy\n"
	  "protocol_state=RCPT\n"
	  "protocol_name=SMTP\n"
	  "helo_name=", 77
	) &&
	obuf_putstr(&netout, &policyd_helo) &&
	obuf_write(&netout, "\nsender=", 8) &&
	obuf_putstr(&netout, &policyd_sender) &&
	obuf_write(&netout, "\nrecipient=", 11) &&
	obuf_putstr(&netout, email_addr) &&
	obuf_write(&netout, "\nclient_address=", 16) &&
	obuf_puts(&netout, client_address) &&
	obuf_write(&netout, "\n\n", 2) &&
	obuf_flush(&netout)
      ) {
        while (ibuf_getstr(&netin, &line, LF)) {
	  str_rstrip(&line);
	  if (line.len == 0) break;
	  if (str_starts(&line, "action=")) {
	    action = line.s + 7;
	    /* get lngth of action string */
	    if ((tmp = strchr(action, ' ')) == 0) {
	      size = strlen(action);
	    }
	    else {
	      size = tmp - action;
	      while (*tmp == ' ') ++tmp;
	    }
	    if (strncasecmp("REJECT", action, size) == 0)
	      resp_policyd.number = 550;
	    else if (strncasecmp("DEFER_IF_PERMIT", action, size) == 0)
	      resp_policyd.number = 451;
	    else if (
	      (*action == '4' || *action == '5') &&
	      size == 3 && (ul = strtoul(action, 0, 10)) >= 400
	    )
	      resp_policyd.number = (unsigned int)ul;

	    if (resp_policyd.number) {
	      if ((size = strlen(tmp)))
	        tmp[size-1] = '\0';	/* cut LF */
	      resp_policyd.message =
		*tmp ? tmp :
		  resp_policyd.number < 500 ?
		    POLICYD_DEFER_MSG : POLICYD_REJECT_MSG;
	      r = &resp_policyd;
	    }
	  }
	}
      }
      obuf_close(&netout);
    }
    else
      r = &resp_oom;
    ibuf_close(&netin);
  }
  else
    r = &resp_oom;

  close(sock);
  return r;
}

static const response *message_end(void)
{
  if (!policyd_lookup_mode) return 0;
  str_copys(&policyd_sender, "");	/* forget any previous sender */
  have_sender = 0;
  return 0;
}

struct plugin plugin = {
  .version     = PLUGIN_VERSION,
  .init        = init,
  .helo        = helo,
  .reset       = reset,
  .sender      = sender,
  .recipient   = recipient,
  .message_end = message_end,
};
